from __future__ import annotations

import logging
import multiprocessing
from os import listdir
from os.path import isfile, join
from pathlib import Path

import pandas as pd
from pandas import DataFrame

from ..utils import multiprocessing_utils, s3_utils
from .load_s3 import list_bucket_prefix_suffix_contains_s3

logger = logging.getLogger(__name__)


# TODO: v1.0 consider renaming function so it isn't 'load'. Consider instead 'load_pd', or something more descriptive.
# TODO: Add full docstring
def load(
    path: str | Path | list[str | Path],
    delimiter=None,
    encoding="utf-8",
    columns_to_keep=None,
    dtype=None,
    header=0,
    names=None,
    format=None,
    nrows=None,
    skiprows=None,
    usecols=None,
    low_memory=False,
    converters=None,
    worker_count=None,
    multiprocessing_method="forkserver",
) -> DataFrame:
    if isinstance(path, Path):
        path = str(path)
    if isinstance(path, list):
        path = [str(p) if isinstance(p, Path) else p for p in path]
        return _load_multipart(
            paths=path,
            delimiter=delimiter,
            encoding=encoding,
            columns_to_keep=columns_to_keep,
            dtype=dtype,
            header=header,
            names=names,
            format=format,
            nrows=nrows,
            skiprows=skiprows,
            usecols=usecols,
            low_memory=low_memory,
            converters=converters,
            worker_count=worker_count,
            multiprocessing_method=multiprocessing_method,
        )
    if format is not None:
        pass
    elif path[-1] == "/" and s3_utils.is_s3_url(path):  # and path[:2] == 's3'
        format = "multipart_s3"
    elif path[-1] == "/" and not s3_utils.is_s3_url(path):  # and path[:2] != 's3'
        format = "multipart_local"
    elif ".parquet" in path or ".pq" in path or path[-1] == "/":
        format = "parquet"
    else:
        format = "csv"
        if delimiter is None:
            if path.endswith(".tsv"):
                delimiter = "\t"
                logger.debug(
                    f"File delimiter for {path} inferred as '\\t' (tab). If this is incorrect, please manually load the data as a pandas DataFrame."
                )
            else:
                delimiter = ","
                logger.debug(
                    f"File delimiter for {path} inferred as ',' (comma). If this is incorrect, please manually load the data as a pandas DataFrame."
                )

    if format == "multipart_s3":
        bucket, prefix = s3_utils.s3_path_to_bucket_prefix(path)
        return _load_multipart_s3(
            bucket=bucket,
            prefix=prefix,
            columns_to_keep=columns_to_keep,
            dtype=dtype,
            worker_count=worker_count,
            multiprocessing_method=multiprocessing_method,
        )  # TODO: Add arguments!
    elif format == "multipart_local":
        paths = [join(path, f) for f in listdir(path) if (isfile(join(path, f))) & (f.startswith("part-"))]
        return _load_multipart(
            paths=paths,
            delimiter=delimiter,
            encoding=encoding,
            columns_to_keep=columns_to_keep,
            dtype=dtype,
            header=header,
            names=names,
            format=None,
            nrows=nrows,
            skiprows=skiprows,
            usecols=usecols,
            low_memory=low_memory,
            converters=converters,
            worker_count=worker_count,
            multiprocessing_method=multiprocessing_method,
        )
    elif format == "parquet":
        df = pd.read_parquet(path, columns=columns_to_keep)
        column_count_full = len(df.columns)
    elif format == "csv":
        df = pd.read_csv(
            path,
            converters=converters,
            delimiter=delimiter,
            encoding=encoding,
            header=header,
            names=names,
            dtype=dtype,
            low_memory=low_memory,
            nrows=nrows,
            skiprows=skiprows,
            usecols=usecols,
        )
        column_count_full = len(list(df.columns.values))
        if columns_to_keep is not None:
            df = df[columns_to_keep]
    else:
        raise Exception("file format " + format + " not supported!")

    row_count = df.shape[0]

    column_count_trimmed = len(list(df.columns.values))

    logger.log(
        20,
        "Loaded data from: "
        + str(path)
        + " | Columns = "
        + str(column_count_trimmed)
        + " / "
        + str(column_count_full)
        + " | Rows = "
        + str(row_count)
        + " -> "
        + str(len(df)),
    )
    return df


def _load_multipart_child(chunk):
    (
        path,
        delimiter,
        encoding,
        columns_to_keep,
        dtype,
        header,
        names,
        format,
        nrows,
        skiprows,
        usecols,
        low_memory,
        converters,
    ) = chunk
    df = load(
        path=path,
        delimiter=delimiter,
        encoding=encoding,
        columns_to_keep=columns_to_keep,
        dtype=dtype,
        header=header,
        names=names,
        format=format,
        nrows=nrows,
        skiprows=skiprows,
        usecols=usecols,
        low_memory=low_memory,
        converters=converters,
    )
    return df


def _load_multipart(
    paths,
    delimiter=",",
    encoding="utf-8",
    columns_to_keep=None,
    dtype=None,
    header=0,
    names=None,
    format=None,
    nrows=None,
    skiprows=None,
    usecols=None,
    low_memory=False,
    converters=None,
    worker_count=None,
    multiprocessing_method="forkserver",
):
    cpu_count = multiprocessing.cpu_count()
    workers = int(round(cpu_count))
    if worker_count is not None:
        if worker_count <= workers:
            workers = worker_count

    logger.log(15, "Load multipart running pool with " + str(workers) + " workers...")

    full_chunks = [
        [
            path,
            delimiter,
            encoding,
            columns_to_keep,
            dtype,
            header,
            names,
            format,
            nrows,
            skiprows,
            usecols,
            low_memory,
            converters,
        ]
        for path in paths
    ]

    df_list = multiprocessing_utils.execute_multiprocessing(
        workers_count=workers,
        transformer=_load_multipart_child,
        chunks=full_chunks,
        multiprocessing_method=multiprocessing_method,
    )

    df_combined = pd.concat(df_list, axis=0, ignore_index=True)

    column_count = len(list(df_combined.columns.values))
    row_count = df_combined.shape[0]

    logger.log(20, "Loaded data from multipart file | Columns = " + str(column_count) + " | Rows = " + str(row_count))
    return df_combined


def _load_multipart_s3(
    bucket,
    prefix,
    columns_to_keep=None,
    dtype=None,
    worker_count=None,
    multiprocessing_method="forkserver",
):
    if prefix[-1] == "/":
        prefix = prefix[:-1]
    prefix_multipart = prefix + "/part-"
    # exclude_contains=['/'] to disallow loading files like 's3://bucket/prefix/part-directory/some_file.abc'
    files = list_bucket_prefix_suffix_contains_s3(bucket=bucket, prefix=prefix_multipart, exclude_contains=["/"])
    paths_full = [s3_utils.s3_bucket_prefix_to_path(bucket=bucket, prefix=file, version="s3") for file in files]

    df = load(
        path=paths_full,
        columns_to_keep=columns_to_keep,
        dtype=dtype,
        worker_count=worker_count,
        multiprocessing_method=multiprocessing_method,
    )
    return df



import os
import platform
from pathlib import Path, PurePosixPath, PureWindowsPath


class PathConverter:
    """Util class to convert a given path to a path to the corresponding OS"""

    @staticmethod
    def _is_windows():
        return platform.system() == "Windows"

    @staticmethod
    def _is_absolute(path: str) -> bool:
        return PureWindowsPath(path).is_absolute() or PurePosixPath(path).is_absolute()

    @staticmethod
    def _validate_path(path: str):
        assert not PathConverter._is_absolute(path), (
            "It is ambiguous on how to convert an absolute path. Please provide a relative path instead"
        )

    @staticmethod
    def to_windows(path: str) -> str:
        PathConverter._validate_path(path)
        return str(PathConverter._to_windows(path))

    @staticmethod
    def _to_windows(path: str) -> PureWindowsPath:
        return PureWindowsPath(PurePosixPath(path))

    @staticmethod
    def to_posix(path: str) -> str:
        PathConverter._validate_path(path)
        return str(PathConverter._to_posix(path))

    @staticmethod
    def _to_posix(path: str) -> PurePosixPath:
        return PurePosixPath(PureWindowsPath(path))

    @staticmethod
    def to_current(path: str) -> str:
        return PathConverter.to_windows(path) if PathConverter._is_windows() else PathConverter.to_posix(path)

    @staticmethod
    def os_path_sep() -> str:
        return os.path.sep

    # v0.9 FIXME: Avoid calling this as much as possible
    #  Refactor code so that calling this is not necessary
    @staticmethod
    def to_relative(path: str) -> str:
        if path == "":
            return path
        if not PathConverter._is_absolute(path=path):
            return path
        os_path_sep = PathConverter.os_path_sep()
        path_relative = os.path.relpath(path)
        if path.endswith(os_path_sep):
            if not path_relative.endswith(os_path_sep):
                path_relative += os_path_sep
        return path_relative

    @staticmethod
    def to_absolute(path: str) -> str:
        if path == "":
            return path
        if PathConverter._is_absolute(path=path):
            return path
        path_absolute = str(Path(path).resolve())
        os_path_sep = PathConverter.os_path_sep()
        if path.endswith(os_path_sep):
            if not path_absolute.endswith(os_path_sep):
                path_absolute += os_path_sep
        return path_absolute



from __future__ import annotations

import os
from urllib.parse import urlparse, urlunparse
from urllib.request import Request, urlopen


def _read_bytes_from_s3(uri: str) -> bytes:
    import boto3

    p = urlparse(uri)
    bucket = p.netloc
    key = p.path.lstrip("/")
    if not bucket or not key:
        raise ValueError(f"Invalid s3 uri: {uri}")

    obj = boto3.client("s3").get_object(Bucket=bucket, Key=key)
    return obj["Body"].read()


def _read_bytes_from_http(uri: str, *, timeout_s: float = 30.0) -> bytes:
    # Basic urllib implementation; no extra deps.
    # Add a UA to avoid some CDNs rejecting default python-urllib.
    req = Request(uri, headers={"User-Agent": "AutoGluon/PresetsLoader"})
    with urlopen(req, timeout=timeout_s) as r:
        return r.read()


def load_preset_dict_from_location(location: str) -> dict:
    """
    Load a preset dict from YAML at:
      - local path
      - s3://bucket/key
      - http(s)://...

    Fragment support:
      - .../presets.yaml#fast -> returns YAML['fast'] (must be dict)
      - no fragment -> returns the YAML top-level dict (backwards compatible)
    """
    p = urlparse(location)
    fragment = p.fragment or None

    # Strip fragment for retrieval
    p_no_frag = p._replace(fragment="")
    location_no_frag = urlunparse(p_no_frag)

    if p.scheme in ("", "file"):
        path = location_no_frag if p.scheme == "" else p_no_frag.path
        if not os.path.exists(path):
            raise FileNotFoundError(path)
        with open(path, "rb") as f:
            data = f.read()
    elif p.scheme == "s3":
        data = _read_bytes_from_s3(location_no_frag)
    elif p.scheme in ("http", "https"):
        data = _read_bytes_from_http(location_no_frag)
    else:
        raise ValueError(f"Unsupported preset URI scheme {p.scheme!r} in {location!r}")

    import yaml

    loaded = yaml.safe_load(data)
    if loaded is None:
        loaded = {}
    if not isinstance(loaded, dict):
        raise TypeError(f"Preset YAML must be a dict at top-level, got {type(loaded)} from {location!r}")

    if fragment is not None:
        if fragment not in loaded:
            raise KeyError(
                f"Preset fragment {fragment!r} not found in {location_no_frag!r}. "
                f"Available keys: {list(loaded.keys())}"
            )
        selected = loaded[fragment] or {}
        if not isinstance(selected, dict):
            raise TypeError(f"Preset {fragment!r} in {location_no_frag!r} must be a dict, got {type(selected)}")
        return selected

    return loaded



import functools
import logging
from typing import Dict
from urllib.parse import urlparse

from .presets_io import load_preset_dict_from_location

logger = logging.getLogger(__name__)


def unpack(g, *other_args):
    """
    Used to pass *args and **kwargs to a function g prior to entering function f.

    Examples
    --------

    >>> def g(a=0, **kwargs):
    >>>     kwargs['b'] = a + 1
    >>>     return kwargs
    >>>
    >>> @unpack(g)
    >>> def f(**kwargs):
    >>>     print(kwargs)
    >>>
    >>> f(a=2)  # kwargs is now the output of g(a=2), which is {'b': 3}
    >>> f(c=4)  # kwargs is now the output of g(c=4), which is {'b': 1, 'c': 4}
    """

    def _unpack_inner(f):
        @functools.wraps(f)
        def _call(*args, **kwargs):
            gargs, gkwargs = g(*other_args, *args, **kwargs)
            return f(*gargs, **gkwargs)

        return _call

    return _unpack_inner


def _looks_like_preset_location(s: str) -> bool:
    """
    Return True only if `s` clearly refers to a file or URL.
    Conservative by design to preserve old error behavior.
    """
    # Explicit schemes
    parsed = urlparse(s)
    if parsed.scheme in {"s3", "http", "https", "file"}:
        return True

    # Local file path heuristics
    if s.endswith((".yaml", ".yml")):
        return True

    # Relative/absolute paths
    if s.startswith(("./", "../", "/")):
        return True

    return False


def _resolve_preset_str(
    preset_og: str,
    preset_dict: Dict[str, dict],
    presets_alias: Dict[str, str] | None,
) -> dict:
    # 1) Built-in preset
    preset = preset_dict.get(preset_og)
    if preset is not None:
        return preset

    # 2) Alias
    if presets_alias is not None:
        mapped = presets_alias.get(preset_og)
        if mapped is not None:
            logger.log(20, f"Preset alias specified: '{preset_og}' maps to '{mapped}'.")
            preset = preset_dict.get(mapped)
            if preset is not None:
                return preset

    # 3) Only try YAML loading if it *looks like* a path / URL
    if _looks_like_preset_location(preset_og):
        try:
            loaded = load_preset_dict_from_location(preset_og)
        except Exception as e:
            raise ValueError(f"Failed to load preset from location {preset_og!r}: {e}") from e

        logger.log(20, f"Loaded presets from {preset_og!r}: keys={list(loaded.keys())}")
        return loaded

    # 4) Otherwise: ORIGINAL error behavior
    valid_presets = list(preset_dict.keys())

    raise ValueError(f"Preset '{preset_og}' was not found. Valid presets: {sorted(set(valid_presets))}")


def _apply_presets(preset_dict: Dict[str, dict], presets_alias: Dict[str, str] = None, *args, **kwargs):
    """
    Pair with `unpack` to alter input arguments with preset values.

    Parameters
    ----------
    preset_dict : Dict[str, dict]
        Dictionary of preset keys that map to dictionaries of key-word values.
    presets_alias : Dict[str, str], optional
        Dictionary of aliases of the presets in preset_dict.
        Aliases will be remapped to the original preset in preset_dict.
    presets : str or list, optional
        List of preset keys (and/or aliases) to apply.
        If str, then it is converted to a 1 element long list.
        presets are applied from first-to-last.
        If a key-word is specified in multiple presets in the list, the value will be set to the value of the last preset with that key-word.
    *args, **kwargs:
        The original args and kwargs (including presets as a potential kwarg).
        args and kwargs take priority over presets, and if specified in the input will not be overwritten.
        Presets will add new key-values to kwargs if the key did not previously exist.

    Returns
    -------
    (*args, **kwargs) with kwargs updated based on specified presets.
    """
    presets = kwargs.get("presets", None)
    if presets is None:
        return args, kwargs

    if not isinstance(presets, list):
        presets = [presets]

    preset_kwargs = {}
    for preset in presets:
        if isinstance(preset, str):
            preset_dict_resolved = _resolve_preset_str(preset, preset_dict, presets_alias)
            preset = preset_dict_resolved

        if isinstance(preset, dict):
            for key, val in preset.items():
                preset_kwargs[key] = val
        else:
            raise TypeError(
                f"Preset of type {type(preset)} was given, but only presets of type [dict, str] are valid."
            )

    # args/kwargs win over presets
    for key, val in preset_kwargs.items():
        if key not in kwargs:
            kwargs[key] = val

    return args, kwargs


def apply_presets(preset_dict: Dict[str, dict], presets_alias: Dict[str, str] = None):
    """Used as a decorator"""
    return unpack(_apply_presets, preset_dict, presets_alias)
